from .gileum import BaseGileum


class MockGileum(BaseGileum):
    """Mock gileum class for testing.

    Attributes:
        test_name (str): Test name to be conducting.
        developer_name (str): Developer name of the package.
        current_time (float): The current Unix time.
    """
    test_name: str
    developer_name: str
    current_time: float



from importlib.machinery import ModuleSpec
import importlib.util
import inspect
import os
from pathlib import Path
import sys
from typing import List, Optional

from .gileum import BaseGileum
from .manager import _get_glm_manager


GILEUM_FILE_SUFFIX = ".glm.py"


def _convert2relative_path(path: str) -> str:
    return str(Path(path).relative_to(os.getcwd()))


def _import_directly(file: str) -> ModuleSpec:
    if os.path.isabs(file):
        file = _convert2relative_path(file)
    mod_name = file.replace(os.sep, ".")
    spec = importlib.util.spec_from_file_location(mod_name, file)
    mod = importlib.util.module_from_spec(spec)
    sys.modules[mod_name] = mod
    spec.loader.exec_module(mod)
    return mod


def _search_glm_from_mod(mod: ModuleSpec) -> List[BaseGileum]:
    f_predicate = lambda obj: isinstance(obj, BaseGileum)
    return [val for _, val in inspect.getmembers(mod, f_predicate)]


def _parse_suffix(suffix: Optional[str]) -> str:
    return GILEUM_FILE_SUFFIX if suffix is None else suffix


def list_glmfiles(
    dir: str,
    join: bool = True,
    suffix: Optional[str] = None
) -> List[str]:
    suffix = _parse_suffix(suffix)
    files = filter(lambda f: f.endswith(suffix), os.listdir(dir))
    if join:
        files = map(lambda f: os.path.join(dir, f), files)

    return list(files)


def load_glms_at(
    file: str,
    name: Optional[str] = None,
    suffix: Optional[str] = None,
) -> None:
    suffix = _parse_suffix(suffix)
    if not os.path.basename(file).endswith(GILEUM_FILE_SUFFIX):
        raise ValueError

    mod = _import_directly(file)
    glms = _search_glm_from_mod(mod)

    manager = _get_glm_manager()
    for glm in glms:
        if name is None:
            manager._set_glm(glm)
        elif glm.__glm_name__ == name:
            manager._set_glm(glm)


def load_glms_in(
    dir: str,
    name: Optional[str] = None,
    suffix: Optional[str] = None
) -> None:
    for gilfile in list_glmfiles(dir, suffix=suffix):
        load_glms_at(gilfile, name=name, suffix=suffix)



from typing import Literal

from pydantic import BaseModel


class BaseGileum(BaseModel):

    # NOTE
    #   This field is used as a part in order to identify cerntain gileum
    #   object in a glm file. The value may be used as a flag to determine
    #   application's behavior.
    glm_name: Literal["main"]



from __future__ import annotations
from copy import deepcopy
from threading import RLock
from typing import (
    Dict,
    Optional,
    Type,
    TypeVar,
)

from .gileum import BaseGileum


Gileum_t = TypeVar("Gileum_t", bound=BaseGileum)


class GileumManager:

    def __init__(self) -> None:
        self.__glms: Dict[str, BaseGileum] = {}

    def _set_glm(self, glm: BaseGileum) -> None:
        if not isinstance(glm, BaseGileum):
            raise TypeError

        if not hasattr(glm, "glm_name"):
            raise AttributeError

        if glm.glm_name in self.__glms:
            return

        self.__glms[glm.glm_name] = glm

    def get_glm(self, name: str, typ: Type[Gileum_t]) -> Gileum_t:
        if name not in self.__glms:
            raise KeyError

        glm = self.__glms.get(name)
        if not isinstance(glm, typ):
            raise TypeError

        return glm

    @property
    def glms(self) -> Dict[str, BaseGileum]:
        return deepcopy(self.__glms)


class SyncGileumManager(GileumManager):

    def __init__(self) -> None:
        super().__init__()

        self.__lock = RLock()

    def _set_glm(self, glm: BaseGileum) -> None:
        with self.__lock:
            super()._set_glm(glm)

    def get_glm(self, name: str, typ: Type[Gileum_t]) -> Gileum_t:
        with self.__lock:
            res = super().get_glm(name, typ)
        return res

    @property
    def glms(self) -> Dict[str, BaseGileum]:
        with self.__lock:
            res = super().glms
        return res


class GileumManagerAlreadySetError(Exception):
    pass


GileumManager_t = TypeVar("GileumManager_t", bound=GileumManager)
__glm_man__: Optional[GileumManager_t] = None


def init_glm_manager(manager: GileumManager_t) -> None:
    global __glm_man__

    # NOTE
    #   __glm_man__ must be a single object within runtime. There are some
    #   reasons for it:
    #
    #   1.  GileumManager object will be in the global scope regardless of
    #       implementation because, in most cases, setting information should
    #       be shared with all objects at runtime. It means the GileumManager
    #       object has its state within the global scope.
    #   2.  If __glm_man__ were resetable, then some objects referencing
    #       __glm_man__ would face the situation that sometimes __glm_man__
    #       had certain gileum, and sometimes __glm_man__ didn't have the one.

    if __glm_man__ is not None:
        raise GileumManagerAlreadySetError
    if not isinstance(manager, GileumManager):
        raise TypeError
    __glm_man__ = manager


def _get_glm_manager() -> GileumManager_t:
    global __glm_man__
    if __glm_man__ is None:
        __glm_man__ = GileumManager()
    return __glm_man__


def get_glm(glm_name: str, typ: Type[Gileum_t]) -> Gileum_t:
    glmman = _get_glm_manager()
    return glmman.get_glm(glm_name, typ)


# NOTE
#   This function is for testing. See the NOTE in `init_glm_manager` to
#   know the reasons for it.
def _reset_glm_manager() -> None:
    global __glm_man__
    __glm_man__ = None
